//=========================================================
//
// アニメーション
//
//=========================================================

#include "Animation.h"
#include "SFML_Utils.h"
#include "SoundPlayer.h"
#include "../Utils/NotXML.h"
#include "../Utils/Utils_Convert.h"
#include "../Utils/Utils_String.h"

using namespace GUI_SFML;

//=========================================================
// コンストラクタ
//=========================================================
Animation::Animation() {

	_position.x = 0.0f;
	_position.y = 0.0f;
	_offset.x = 0.0f;
	_offset.y = 0.0f;
	_isLoop     = true;
	_frameCount = 0;

	_currentFrameIndex = -1;
	_isActive          = false; // 設定を入れると稼動中になる。というか、する。
}
//=========================================================
// デストラクタ
//=========================================================
Animation::~Animation() {

}

//=========================================================
// 設定
//=========================================================
void Animation::Init ( string fileName ) {

	NotXML::Tag root = NotXML().ReadRoot( fileName );

	///// Loop with ChildTags of Root
	for (unsigned int i = 0; i < root.childList.size(); ++i) {

		NotXML::Tag tag = root.childList[i];

		///// ループ指定
		if (tag.name == "Loop") {

			_isLoop = Utils_Convert::String_to_bool ( tag.value );

		///// 画像指定
		} else if (tag.name == "Image") {

			sf::Image   image;
			sf::Texture texture;
			string      label;
			
			for (unsigned int n = 0; n < tag.childList.size(); ++n) {

				NotXML::Tag childTag = tag.childList[n];

				if ( childTag.name == "File" ) {

					image.loadFromFile ( childTag.value );

					string maskString = NotXML::GetAttribute ( childTag, "mask", "");

					if ( maskString != "" ) {

						sf::Color maskColor = SFML_Utils::Convert_string_to_color ( maskString );
						image.createMaskFromColor ( maskColor );
					}

					texture.loadFromImage ( image );
					texture.setSmooth(true);


				} else if ( childTag.name == "Label" ) {
					label = childTag.value;
				}
			}

			_imageList   [ label ] = image;
			_textureList [ label ] = texture;

		///// フレーム指定
		} else if (tag.name == "Frame") {

			AnimationFrame frame;

			for (unsigned int n = 0; n < tag.childList.size(); ++n) {

				NotXML::Tag childTag = tag.childList[n];

				if ( childTag.name == "Label" ) {
					frame.label = childTag.value;
				} else if ( childTag.name == "FrameCount" ) {
					frame.frameCount = Utils_Convert::String_to_Int ( childTag.value );
				} else if ( childTag.name == "NextLabel" ) {
					frame.nextLabel = childTag.value;
				} else if ( childTag.name == "ImageLabel" ) {
					frame.imageName = childTag.value;
				} else if ( childTag.name == "Rect_x" ) {
					frame.rect.left = Utils_Convert::String_to_Int ( childTag.value );
				} else if ( childTag.name == "Rect_y" ) {
					frame.rect.top = Utils_Convert::String_to_Int ( childTag.value );
				} else if ( childTag.name == "Offset_x" ) {
					frame.offsetX = Utils_Convert::String_to_Int ( childTag.value );
				} else if ( childTag.name == "Offset_y" ) {
					frame.offsetY = Utils_Convert::String_to_Int ( childTag.value );
				} else if ( childTag.name == "SetOriginX" ) {
					frame.setOriginX = childTag.value;
				} else if ( childTag.name == "SetOriginY" ) {
					frame.setOriginY = childTag.value;
				} else if ( childTag.name == "ScaleX" ) {
					frame.scaleX = childTag.value;
				} else if ( childTag.name == "ScaleY" ) {
					frame.scaleY = childTag.value;
				} else if ( childTag.name == "Rect_width" ) {
					frame.rect.width = Utils_Convert::String_to_Int ( childTag.value );
				} else if ( childTag.name == "Rect_height" ) {
					frame.rect.height = Utils_Convert::String_to_Int ( childTag.value );
				} else if ( childTag.name == "ReverseX" ) {
					frame.isReverseX = Utils_Convert::String_to_bool ( childTag.value );
				} else if ( childTag.name == "ReverseY" ) {
					frame.isReverseY = Utils_Convert::String_to_bool ( childTag.value );
				} else if ( childTag.name == "AddRotate" ) {
					frame.addRotate = Utils_Convert::String_to_Int ( childTag.value );
				} else if ( childTag.name == "Sound" ) {
					frame.soundFile = childTag.value;
				} else if ( childTag.name == "Color" ) {
					frame.color = childTag.value;
				}
			}

			_frameList.push_back( frame );
		}
	}


	///// 設定できましたよね？
	if ( _frameList.size() ) {

		_frameCount        = 0;
		_currentFrameIndex = 0;
		_isActive = true;

		//StartNewFrame();
	}

}


//=========================================================
// 更新
//=========================================================
void Animation::Update ( sf::RenderWindow &window ) {

	///// 稼動中？
	if ( _isActive ) {

		// 初回？
		if ( _frameCount == 0 && _currentFrameIndex == 0 ) {

			StartNewFrame();

			++_frameCount;

			return;
		}

		++_frameCount;

		// 次のフレームへ？
		if ( _frameCount >= _frameList[_currentFrameIndex].frameCount ) {

			_frameCount = 0;

			///// 次はどのフレーム？ 指定なし
			if ( _frameList[_currentFrameIndex].nextLabel == "" ) {

				// リスト上の次のフレームへ
				++_currentFrameIndex;

				// 終端？
				if ( _currentFrameIndex >= _frameList.size() ) {

					// ループする？
					if ( _isLoop ) {
						_currentFrameIndex = 0;

					// ループしない：終了
					} else {
						_isActive = false;
					}
				}

			///// フレーム指定あり
			} else {
				
				///// 該当のフレームがなければ終了
				_isActive = false;

				///// 該当のフレームを探す
				for ( unsigned int i = 0; i < _frameList.size(); ++i ) {

					// あった
					if (_frameList[i].label == _frameList[_currentFrameIndex].nextLabel) {
						_currentFrameIndex = i;
						_isActive          = true;
						break;
					}
				}
			}


			// ここまできて、まだ終了してないなら、新フレームを表示する準備をする
			if ( _isActive ) {

				StartNewFrame();
			}

		} // 次のフレームへ？



	} // 稼動中？


}

//=========================================================
// 新たなフレームを開始する
//=========================================================
void Animation::StartNewFrame() {

	_offset.x = _frameList[_currentFrameIndex].offsetX;
	_offset.y = _frameList[_currentFrameIndex].offsetY;

	// 該当の画像が登録されてないとか、知りません。
	_sprite.setTexture     ( _textureList[ _frameList[_currentFrameIndex].imageName ] );
	_sprite.setTextureRect ( _frameList[_currentFrameIndex].rect );

	// 回転
	_sprite.rotate ( _frameList[_currentFrameIndex].addRotate );

	// 中心点をセット
	if ( _frameList[_currentFrameIndex].setOriginX != "" ) {
		int originX = Utils_Convert::String_to_Int(_frameList[_currentFrameIndex].setOriginX);
		_sprite.setOrigin ( sf::Vector2f( originX, _sprite.getOrigin().y) );
	}
	if ( _frameList[_currentFrameIndex].setOriginY != "" ) {
		int originY = Utils_Convert::String_to_Int(_frameList[_currentFrameIndex].setOriginY);
		_sprite.setOrigin ( sf::Vector2f( _sprite.getOrigin().x, originY) );
	}

	// 横反転？
	if ( _frameList[_currentFrameIndex].isReverseX ) {
		_sprite.setScale ( -1.0f, 1.0f );
	} else {
		_sprite.setScale ( 1.0f, _sprite.getScale().y );
	}
	// 縦反転？
	if ( _frameList[_currentFrameIndex].isReverseY ) {
		_sprite.setScale ( 1.0f, -1.0f );
	} else {
		_sprite.setScale ( _sprite.getScale().x, 1.0f );
	}

	// スケールX？
	if ( _frameList[_currentFrameIndex].scaleX != "" ) {

		float scaleX = Utils_Convert::String_to_Float ( _frameList[_currentFrameIndex].scaleX  );
		_sprite.scale ( scaleX, 1.0f );
	}
	// スケールY？
	if ( _frameList[_currentFrameIndex].scaleY != "" ) {

		float scaleY = Utils_Convert::String_to_Float ( _frameList[_currentFrameIndex].scaleY  );
		_sprite.scale ( 1.0f, scaleY );
	}

	// 色設定する？
	if ( _frameList[_currentFrameIndex].color != "" ) {
		sf::Color color = SFML_Utils::Convert_string_to_color(_frameList[_currentFrameIndex].color);
		_sprite.setColor(color);
	}

	// 音を鳴らす？
	SoundPlayer::Play ( _frameList[_currentFrameIndex].soundFile );
	
}

//=========================================================
// 描画
//=========================================================
void Animation::Draw   ( sf::RenderWindow &window ) {

	if ( _isActive ) {

		// なんとなく、毎回。
		_sprite.setPosition( _position + _offset );

		window.draw ( _sprite );
	}
}








