//=========================================================
//
// テキストメッセージ欄
//
//=========================================================

#ifndef _TEXT_MESSAGE_WINDOW_
#define _TEXT_MESSAGE_WINDOW_

#include <vector>
#include <string>
#include <SFML/Graphics.hpp>

class IScenarioLogger;

using namespace std;

namespace GUI_SFML {
	
	class ScenarioLogViewer;

	//=========================================================
	// 既読管理インターフェイス。これにそって実装せよ、という思想
	//=========================================================
	class IReadManagerReceptionist {

	public:
		virtual ~IReadManagerReceptionist(){}
		virtual int  GetReadFinishedLength () = 0;     // 何文字目まで読み終えてますか？
		virtual void Finished () = 0;                  // 全部読み終えた
		virtual void Finished ( int charLength ) = 0;  // ｎ文字目まで読み終えた
		virtual bool IsSkip() = 0;                     // スキップ中かどうか
		virtual void SkipEnd() = 0;                    // 既読スキップを終了します！
		virtual void SkippingNow(bool arg) = 0;        // 現在既読スキップ中です。
		virtual bool IsWannaSkip() = 0;                // 今、既読（ならば）スキップすべき？
	};

	class TextChar;
	class Animation;

	class TextMessageWindow : public IReadManagerReceptionist {

	public:

		//=========================================================
		// デザイン情報
		//=========================================================
		struct Design {

			///// 表示 / 非表示
			bool isShown;

			sf::Vector2f position;
			sf::Vector2f size;           // 横幅も、縦幅も、サイズ？ 行数ではなく？
			sf::Vector2f offset;         // 文字の表示を開始する位置のオフセット
			int   fontSize;              // 文字のサイズ
			float rowHeight;             // 行間（間というか、１行の高さ）
			sf::Color fontColor;         // 文字色
			sf::Color fontColor_reRead;  // 文字色：読み返し行

			string     backImageFileName;

			sf::Vector2f speakerOffset;
			float        speakerOffset_centering;
			string       speakerNameDisplayType;
			string       speakFrom;  // 話者名挿入時の目印
			string       speakEnd;   // セリフ終了の目印

			string speakerNameBackImage;
			sf::Vector2f speakerBackImageOffset;

			vector<string> speakerNameBackImage_wide;
			
			// カーソル速度
			int charShowCountMax;
			
			// 画像を使わずに矩形を背景にする場合の情報
			sf::Color frameFillColor;
			sf::Color frameOutlineColor;
			int frameOutlineThickness;
			
		};
		

		//=========================================================
		// IReadManagerReceptionist が NULLのときに使うデフォルト処理
		//=========================================================
		int  GetReadFinishedLength () { return 0; }
		void Finished () {}
		void Finished ( int charLength ) {}
		bool IsSkip() { return false; }
		void SkipEnd() {}
		void SkippingNow(bool arg) {}
		bool IsWannaSkip() { return false; }

		//=========================================================
		// 特殊制御用文字
		//=========================================================
		static string CONTROL_CHAR_NEW_LINE;   // 改行
		static string CONTROL_CHAR_NEW_PAGE;   // 改ページ
		static string CONTROL_CHAR_WAIT_CLICK; // クリック待ち
		
		static void Set_CONTROL_CHAR_NEW_LINE   (string tag) { CONTROL_CHAR_NEW_LINE = tag; }
		static void Set_CONTROL_CHAR_NEW_PAGE   (string tag) { CONTROL_CHAR_NEW_PAGE = tag; }
		static void Set_CONTROL_CHAR_WAIT_CLICK (string tag) { CONTROL_CHAR_WAIT_CLICK = tag; }
		
		//=========================================================
		// 予約文字
		//=========================================================
		static string ALIAS_START; // エイリアス開始タグ
		static string ALIAS_END;   // エイリアス終了タグ
		
		//=========================================================
		// 話者名表示方式
		//=========================================================
		//const static string SPEAKER_NAME_DISPLAY_TYPE_NON; // 非表示 // 話者名なしをもって非表示とする。話者名が指定されているときに、いかに表示するか？ という話なのであって
		const static string SPEAKER_NAME_DISPLAY_TYPE_OUT; // 外側（名前欄に表示する）
		const static string SPEAKER_NAME_DISPLAY_TYPE_IN;  // 内側（行の冒頭に挿入する）

		TextMessageWindow ( IReadManagerReceptionist *readManagerReceptionist );
		~TextMessageWindow();
		
		void Update( sf::RenderWindow &window );
		void Draw  ( sf::RenderWindow &window );

		void Clear();

		void AddText ( string text );
		void AddText ( string text, sf::Color fontColor );
		void AddText_asFinishedText ( string text );
		void AddText_asFinishedText ( string text, sf::Color fontColor );
		//void AddText_outOfHistory ( string text );
		//void AddText_outOfHistory ( string text, sf::Color fontColor );
		bool IsAllCharShown();

		void SetPosition ( sf::Vector2f position ) { _position = position; }
		void SetPositionX ( int x ) { _position.x = (float)x; }
		void SetPositionY ( int y ) { _position.y = (float)y; }

		void SetOffset ( sf::Vector2f offset ) { _offset = offset; }
		void SetOffsetX ( int x ) { _offset.x = (float)x; }
		void SetOffsetY ( int y ) { _offset.y = (float)y; }

		void SetWidth  ( int value ) { _size.x = value; }
		void SetHeight ( int value ) { _size.y = value; }

		void SetFontSize  ( int value ) { _fontSize  = value;}
		void SetRowHeight ( int value ) { _rowHeight = (float)value;}
		void SetFontColor ( sf::Color color ) { _fontColor = color; }
		void SetFontColor_reRead ( sf::Color color ) { _fontColor_reRead = color; }

		void SetBackImage ( string fileName );

		// カーソル速度（次の文字が表示されるまで待つフレーム数）：値が小さいほど速い。
		void SetCursorSpeed ( int value ) { _charShowCountMax = value; }  

		void SetFrameFillColor        ( sf::Color color ) { _frame.setFillColor ( color ); }
		void SetFrameOutlineColor     ( sf::Color color ) { _frame.setOutlineColor ( color ); }
		void SetFrameOutlineThickness ( int value )       { _frame.setOutlineThickness ( value ); }

		void SetClickWaitCorsor ( string fileName );

		///// 話者名をセット
		void SetSpeakerName ( string speakerName );
		void SetSpeakerNameOffset ( sf::Vector2f pos ) { _speakerOffset = pos; }
		void SetSpeakerNameOffsetX ( float x) { _speakerOffset.x = x; }
		void SetSpeakerNameOffsetY ( float y) { _speakerOffset.y = y; }
		void SetSpeakerNameOffsetCentering ( float value) { _speakerOffset_centering = value; } // 中央寄せをする場合の「右端」までの長さ。0なら左詰め。

		void SetSpeakerNameBackImage ( string fileName );
		void SetSpeakerNameBackImage_wide ( string fileName );

		void SetSpeakerNameBackImage_wide ( vector<string> fileNameList );
		void ClearSpeakerNameBackImage_wide ();
		
		
		void SetSpeakerNameBackImageOffsetX ( float x) { _speakerBackImageOffset.x = x; }
		void SetSpeakerNameBackImageOffsetY ( float y) { _speakerBackImageOffset.y = y; }
		
		void   SetSpeakerNameDisplayType ( string type );
		string GetSpeakerNameDisplayType () {return _speakerNameDisplayType;}
		void   SetSpeakFrom ( string arg ) { _speakFrom = arg; }
		string GetSpeakFrom () { return _speakFrom; }
		void   SetSpeakEnd ( string arg ) { _speakEnd = arg; } // TODO _speakEnd は使っていないのではありませんでした？
		string GetSpeakEnd () { return _speakEnd; }
		
		///// 文字列のエイリアスを登録
		static void RegistTextAlias ( string alias, string value );
		static void DeleteTextAlias ( string alias );
		
		static void RegistProtagonist ( string name );
		static void ShowProtagonistName ();
		static void HideProtagonistName ();
		
		
		///// 全部表示し終わっているか？ クリック待ちでもなく（試合中のアクターのセリフとして、「完了」しているかどうかの判定が目的）
		bool IsAllFinished();
		
		//=========================================================
		// テキスト内のエイリアスを置換
		//=========================================================
		static void ReplaceWithAlias ( string &text );
		
		///// デザイン情報を取得/セット
		Design GetDesign ();
		void SetDesign ( Design design);

		//=========================================================
		// セーブ/ロード用インターフェイス
		//=========================================================
		struct SaveData {

			/*
				ウィンドウそのものの定義・設定項目はセーブ/ロード対象にはしません。
				あらかじめ定義しておいて、そこへセーブデータを流し込む。
			*/
			vector < pair <string, string > > addTextHistory;
			bool isWaitingClick;           // クリック待ち？
			int  textCharListLength;       // 何文字目まで表示されているかを覚えておく。

			vector< string > reReadRowList; // 読み返し行リスト
			
			/*
				AddText の履歴を保存しておいてはどうか？
				改ページのときに、クリア。
				渡された引数。文字列。色、のリスト。
				色はこの時点で文字列化しておくか。
				
				どこまで表示されているか？
				「実際に表示されている分」の文字数を、数字で持っておく。

			*/

			// セーブ/ロード対象ではないけど、ロード時にクリアしてげるとよさそうなもの
			//int _charShowCount; // １文字ずつ表示するためのトリック
			//int _charShowCountMax;
			
			
			bool isShown;
			string speakerName;
			
			string addText_latest;
			
			int stepped; // 何文字目まで表示していたか？
			

			SaveData() {
				isShown = false;
				speakerName = "";
				addText_latest = "";
				stepped = 0;
			}

		};

		SaveData GetSaveData();
		void Load ( SaveData &saveData );
		
		
		///// 読み返しモード中？
		bool IsScenarioLogMode() { return _isReReadMode; }
		
		
		///// コピーというか、同じ設定のインスタンスを取得
		TextMessageWindow* GetSameSetuppedInstance();


		void SetShown( bool arg ) { _isShown = arg; }
		bool IsShown ()           { return _isShown; }
		
		///// 読み返しログ処理用オブジェクト
		void SetScenarioLogger (IScenarioLogger *obj);
		

		///// カメラの位置を気にするのをヤメる：画面揺れエフェクトの間だけ、という意図
		bool _notCareOfCameraPos;
		
		
		static std::vector<string> GetProtagonistsNames() { return _protagonists; }

	private:
		
		
		///// 表示 / 非表示
		bool _isShown;

		sf::Vector2f _position;
		sf::Vector2f _size;           // 横幅も、縦幅も、サイズ？ 行数ではなく？
		sf::Vector2f _offset;         // 文字の表示を開始する位置のオフセット
		int   _fontSize;              // 文字のサイズ
		float _rowHeight;             // 行間（間というか、１行の高さ）
		sf::Vector2f _cursorPosition; // カーソル位置
		sf::Color _fontColor;         // 文字色
		sf::Color _fontColor_reRead;  // 文字色：読み返し行

		// フレームを背景に使う？
		bool _isFrameOn;
		sf::RectangleShape _frame;

		// 画像を背景に使う？
		bool _isBackImageOn;
		string      _backImageFileName;
		sf::Image   _image;
		sf::Texture _texture;
		sf::Sprite  _sprite;

		vector<TextChar*> _textCharList;          // 実際に表示されている分
		vector<TextChar*> _textCharList_Reserved; // まだ表示されてない分

		///// 話者名
		string _speakerName;
		//sf::Image   _speakerName_image;
		//sf::Texture _speakerName_texture;
		//sf::Sprite  _speakerName_sprite;
		sf::Text    _speakerName_sprite;
		sf::Vector2f _speakerOffset;
		float        _speakerOffset_centering;
		string       _speakerNameDisplayType;
		string       _speakFrom;  // 話者名挿入時の目印
		string       _speakEnd;   // セリフ終了の目印

		// 話者名枠画像
		bool _isSpeakerNameBackImageOn;
		string       _speakerNameBackImageFileName;
		sf::Vector2f _speakerBackImageOffset;
		sf::Image    _speakerNameBackImage_image;
		sf::Texture  _speakerNameBackImage_texture;
		sf::Sprite   _speakerNameBackImage_sprite;

		//string       _speakerNameBackImageFileName_wide;
		//sf::Texture  _speakerNameBackImage_wide_texture;
		//sf::Sprite   _speakerNameBackImage_wide_sprite;
		
		vector<string>       _speakerNameBackImageFileName_wide;
		vector<sf::Texture>  _speakerNameBackImage_wide_texture;
		vector<sf::Sprite>   _speakerNameBackImage_wide_sprite;
		
		// 現在、どの名札プレートを使えばいいのか？ : 負の数 でデフォルト。0以降でワイド版
		int _currentamePlateNum;
		
		
		///// 文字列のエイリアス
		static std::map<string, string> _textAliasList;
		
		///// 主人公扱いの話者名
		static std::vector<string> _protagonists;
		
		///// 主人公名を話者名欄に表示するかどうか？
		static bool _isProtagonistShown;
		

		
		// セーブ/ロード用
		/*
			pair < 文字列, 色 > のリスト
			改ページのタイミングでクリアされる。
		*/
		vector < pair <string, string > > _addTextHistory;
		
		// addTextのタイミング、すなわち、Textタグ単位、ということ。：これをセーブ対象としよう。
		string _addText_latest;
		
		// 最後の addText以降、何文字表示したか？
		int _charCount_sinceLastAddText;
		
		
		

		int _charShowCount; // １文字ずつ表示するためのトリック
		int _charShowCountMax;

		bool _isWaitingClick;           // クリック待ち？
		Animation *_waitingClickCursor; // クリック待ちのときに表示するアイコン

		bool _isTemporaryQuick;         // 文字が流れている最中に左クリックしたときのアレ。

		// 既読管理インターフェイス
		IReadManagerReceptionist *_readManagerReceptionist;

		// 何文字目まで既読か？ AddTextの時点で、受け取っておく。
		int _readFinishedLength; 

		int _mouseStateL; // マウスの状態
		int _mouseStateR; // マウスの状態

		///// 読み返しモード関連
		vector< vector< TextChar* > > _reReadRowList;        // 読み返し行リスト
		vector< vector< TextChar* > > _reReadRowListCurrent; // 読み返し行リスト。現時点で表示している分
		bool _isReReadMode;      // 今、読み返しモード？
		int  _reReadRangeFrom;   // 読み返し行を表示する範囲
		int  _reReadRangeTo;     // 読み返し行を表示する範囲
		int  _reReadLimitRowNum; // 読み返せる行数の上限
		
		///// 読み返しモードにする？ ：マウスの操作を監視して、モードを切り替える（読み返しモードじゃないときに呼ぶ）
		void ToReReadModeIfNeccesary ( sf::RenderWindow &window );

		///// 通常モードにする？：マウスの操作を監視して、モードを切り替える（読み返しモードのときに呼ぶ）
		void ToNomalModeIfNeccesary ();

		///// 読み返しモードのときの更新
		void Update_ReReadMode ( sf::RenderWindow &window );

		///// 読み返し文字さんたちに位置についてもらう
		void Update_ReReadCharList( sf::RenderWindow &window );

		///// 読み返しモードのときの操作受付
		void ReReadMode_Operation ();

		///// 読み返し用：１行の幅を求める
		//int GetRowHeight ( vector< vector< TextChar* > > &reReadRowList );

		///// 現在の分の読み返し行を求める
		void SetUpCurrentReReadCharList();

		///// 読み返し行をクリアする
		void Release_ReReadCharList();
		
		
		///// 読み返しログ処理用オブジェクト
		IScenarioLogger *_scenarioLogger;
		
		///// 読み返しログビューワー
		ScenarioLogViewer *_scenarioLogViewer;

		//=========================================================
		// 内部定数
		//=========================================================
		static const int MOUSE_STATE_NONE;    // 押していない
		static const int MOUSE_STATE_PRESS;   // 押している
		static const int MOUSE_STATE_RELEASE; // 放したところ

		//=========================================================
		// １文字ずつ表示するトリック
		//=========================================================
		void ShowCharOneByOne( sf::RenderWindow &window );
		
		//=========================================================
		// 改ページ
		//=========================================================
		void NewPage();

		//=========================================================
		// スクロール
		//=========================================================
		void Scroll();

		//=========================================================
		// スクロール時、削った最上行の文字を始末
		//=========================================================
		void PutScrolledRowAway ( vector<TextChar*> &firstRow );

		//=========================================================
		// マウスの状態を監視
		//=========================================================
		void ObserveMouseState ( sf::RenderWindow &window );

		//=========================================================
		// string を １文字ごとに区切る
		//=========================================================
		vector<string> SplitOneByOne ( string text );


		//=========================================================
		// １文字進める
		//=========================================================
		void ProgressOneChar();

		//=========================================================
		// 行に分割する
		//=========================================================
		//vector < vector<TextChar*> > SplitIntoRows( vector<TextChar*> &charList );


		//=========================================================
		// 制御文字？
		//【引数】
		// int &currentIndex      ：制御文字だったら、いい位置まで進めてしまう。
		// sf::Color &currentColor：制御文字だったら、色を変えてしまう。 
		//=========================================================
		bool IsColorTag ( vector<string> &charList, unsigned int &currentIndex, const string &colorTag, sf::Color &currentColor );

		//=========================================================
		// ルビ指定タグ？
		//【引数】
		// int &currentIndex ：タグだったら、いい位置まで進めてしまう。
		// string &ruby      ：ルビ
		//=========================================================
		bool IsRubyTag ( vector<string> &charList, unsigned int &currentIndex, const string &rubyTag, string &ruby );


		bool IsControlTag ( vector<string> &charList, unsigned int &currentIndex, const string &controlTag );

		//=========================================================
		// テキスト追加：処理本体
		//=========================================================
		void AddText_Body ( string text, sf::Color fontColor );
		
		//=========================================================
		// 読み返しログ用の文字列を溜め込んでいく
		//=========================================================
		void StoreCharToScenarioLog ( string c );

		//=========================================================
		// 溜まっている読み返しログを書き込む
		//=========================================================
		void FlushToScenarioLog ();
		
		string _scenarioLogBuffer;
		//bool _isSpeakingRightBefore; // 直前の書き込みがセリフだった：つまり、見た目上、行を空けたい
		bool _isSpeakingStarted;
		bool _isSpeakingEndedRightBefore;
		
		
		
		//=========================================================
		// 禁則処理の文字かどうか？
		//=========================================================
		bool IsHyphenatedChar_notSuitableForTheTopOfLine ( const string &c );
		bool IsHyphenatedChar_notSuitableForTheTailOfLine ( const string &c );
		

	};

};

#endif







