//=========================================================
//
// ゲーム盤管理クラス
//
//=========================================================

#ifndef _GAME_MANAGER_
#define _GAME_MANAGER_

#include <./SFML/Graphics.hpp>
#include <vector>

#include "IEffect.h"
#include "../Utils/NotXML.h"
#include "../AI/AI.h"
#include "../AI/Move.h"
#include "../AI/InnerExpression.h"

#include "../GUI_SFML/ProgressBar.h"
#include "../GUI_SFML/TextMessageWindow.h"

#include "Actor.h"


using namespace std;

class Board;
class CardBox;
class PieceBox;
class Piece;
class Card;
class Grid;
class AI_Manager;
class DelegateBase;
class GameRecord;
class PsychoGage;
class Arrow;
class Image;

class Select;
class InterfereModePanel;
class CurrentTurnNamePanel;
class ScriptEngine;


class GameManager {

public:
	

	//=========================================================
	// 候補手からの選択モードバトルの情報
	//
	// 参加者が持ちつつ、手番の開始時にGamaManager側にコピーし、手番終了時に参加者側にコピーしなおす。
	//=========================================================
	struct ChoiceModeInfo {
		
		bool isActive;            // 候補手からの選択モードであるか？
		int numberOfChoices;      // 表示する選択肢数
		string selectOptionsPos; // 選択肢の表示位置
		vector< pair<int,int> > extraChoicesRange; // ２番目以降の選択肢の取得基準（一番目は必ずAI選定のものを使用する）
		
		vector<AI_Modules::ProgressData> moveListSuggestedByAI; // AIからの候補手リスト
		string playersSelectChoiceNumber; // プレイヤーが選択した選択肢番号。空文字で未選択、ということにする
		
		string phase; // 進行フェイズ
		// 進行フロー：
		//     WAIT_FOR_AI_MOVE          : AIから候補手を待っているところ
		//     WAIT_FOR_PLAYER_DECISION  : 「どれにする？」とプレイヤーの回答を待っているところ
		//     AFTER_PLAYER_DECIDED      : プレイヤーの回答を受け取った後
		//     WAIT_FOR_NEXT_MY_TURN     : 次の自分の手番を待っているところ
		//		
		
		// const でも static でもないですけど、もうどうでもいいや。
		string PHASE_WAIT_FOR_AI_MOVE;
		string PHASE_WAIT_FOR_PLAYER_DECISION;
		string PHASE_AFTER_PLAYER_DECIDED;
		string PHASE_WAIT_FOR_NEXT_MY_TURN;
		
		ChoiceModeInfo () {
			isActive         = false;
			numberOfChoices  = 0;
			selectOptionsPos = "";
			playersSelectChoiceNumber = "";

			PHASE_WAIT_FOR_AI_MOVE         = "WAIT_FOR_AI_MOVE";
			PHASE_WAIT_FOR_PLAYER_DECISION = "WAIT_FOR_PLAYER_DECISION";
			PHASE_AFTER_PLAYER_DECIDED     = "AFTER_PLAYER_DECIDED";
			PHASE_WAIT_FOR_NEXT_MY_TURN    = "WAIT_FOR_NEXT_MY_TURN";

			phase = PHASE_WAIT_FOR_NEXT_MY_TURN;
			
		}
	};
	//=========================================================
	// 口出しモードバトルの情報
	//
	// 参加者が持ちつつ、手番の開始時にGamaManager側にコピーし、手番終了時に参加者側にコピーしなおす。
	//=========================================================
	struct InterferenceModeInfo {
		
		bool isActive;    // 口出しモードであるか？
		int maxTimes;     // 最大回数
		int currentTimes; // 現在の口出し回数：最初は0
		
		AI_Modules::Move moveSuggestedByAI; // AIの指し手案（これでいい？ の確認対象）
		bool isAgreedWithAIMove;            // AIの指し手案が同意されている
		
		string selectOptionsPos; // 選択肢の表示位置
		
		string phase; // 進行フェイズ
		// 進行フロー：
		//     WAIT_FOR_AI_MOVE             : AIから候補手を待っていたところ
		//     WAIT_FOR_PLAYER_CONFIRMATION : 「これでいい？」の質問を表示してプレイヤーの回答を待っているところ
		//     WAIT_FOR_PLAYER_MOVE         : プレイヤーの操作（着手）を待機しているところ
		//     WAIT_FOR_NEXT_MY_TURN        : 次の自分の手番を待っているところ
		//		
		
		// const でも static でもないですけど、もうどうでもいいや。
		string PHASE_WAIT_FOR_AI_MOVE;
		string PHASE_WAIT_FOR_PLAYER_CONFIRMATION;
		string PHASE_WAIT_FOR_PLAYER_MOVE;
		string PHASE_WAIT_FOR_NEXT_MY_TURN;
		
		InterferenceModeInfo () {
			isActive          = false;
			maxTimes          = 0;
			currentTimes      = 0;
			moveSuggestedByAI.action = 0;
			isAgreedWithAIMove = false;
			
			selectOptionsPos = "";

			PHASE_WAIT_FOR_AI_MOVE             = "WAIT_FOR_AI_MOVE";
			PHASE_WAIT_FOR_PLAYER_CONFIRMATION = "WAIT_FOR_PLAYER_CONFIRMATION";
			PHASE_WAIT_FOR_PLAYER_MOVE         = "WAIT_FOR_PLAYER_MOVE";
			PHASE_WAIT_FOR_NEXT_MY_TURN        = "WAIT_FOR_NEXT_MY_TURN";

			phase = PHASE_WAIT_FOR_NEXT_MY_TURN;
		}
	};
	
	
	//=========================================================
	// 試合の参加者
	//=========================================================
	struct Participant {
		
		string id;   // 内部処理用の識別子
		string side; // P1 / P2
		string name; // 表示用の名前
		
		// セーブ/ロード用にAIおよびアクターの設定ファイル名を覚えておく
		string aiFileName;
		string actorFileName;
		
		// セーブロード用にアクターの現在位置と目的位置を覚えておく（セーブ/ロードの時点で）
		int actorGoalPosX, actorGoalPosY;
		int actorCurrentPosX, actorCurrentPosY;
		
		// この参加者の手番である場合の経過ターン数
		// 初期値を -1 としておき、「開始直後でまだ何もしていない」を表せるのかもしれない。
		long turnFrameCount;
		
		AI_Modules::AI *ai;
		Actor *actor;
		
		InterferenceModeInfo interferenceModeInfo;
		ChoiceModeInfo       choiceModeInfo;
		
		Participant (){
			id    = "";
			side  = "";
			name  = "";
			ai    = NULL;
			actor = NULL;
			turnFrameCount = -1;
			
			aiFileName = "";
			actorFileName = "";
			actorGoalPosX = 0;
			actorGoalPosY = 0;
			actorCurrentPosX = 0;
			actorCurrentPosY = 0;
		}
		
	};

	//=========================================================
	// 手番を管理するデータ
	//=========================================================
	struct OperationData {

		bool isLock;  // 操作、ロックされている？

		// 手番の順番リスト：参加者 participant の id のリスト。１ターン分
		vector<string> turnOrder;
	
		string currentTurn; // 現在の手番（ turnOrder内のid、のはず ）SetTurnOrder した時点で先頭のものにする。
		int currentTurnIndex;

		///// コンストラクタ
		OperationData () {
			isLock = false;
			currentTurnIndex = 0;
			currentTurn = "";
		}
		
		///// 手番を最初の番の人にする
		void SetTurnToFirst() {
			
			currentTurnIndex = 0;
			
			if ( turnOrder.size() >= 1 ) {
				currentTurn = turnOrder[0];
			} else {
				currentTurn = "";
			}
		}
		
		///// 手番を次の人へ進める
		void TurnToNext () {
			
			currentTurnIndex += 1;

			if ( currentTurnIndex >= turnOrder.size()  ) {
				currentTurnIndex = 0;
			}

			currentTurn = turnOrder[currentTurnIndex];
			
		}
	};

	//=========================================================
	// 試合関連のデータ
	//=========================================================
	struct BattleInfo {

		bool isBattlingNow; // 試合中？
		string result;      // 試合結果："" 未定  "P1Win" "P2Win" "Draw"　：これは暫定。
		
		string battleTitle; // 試合名
		
		string resultDescription; // 棋譜の最後の「勝敗結果」として使う文言

		BattleInfo () {
			isBattlingNow = false;
			result = "";
			battleTitle = "";
			resultDescription = "";
		}
	};

	GameManager(string name, ScriptEngine *scriptEngine);
	~GameManager();

	//void Init ();
	void SetUpBoard ( int width, int height );

	void Update           ( sf::RenderWindow &window );
	void Update_Operation ();
	void Update_Battle    ();
	void Update_AstralProjection ();

	void Draw ( sf::RenderWindow &window );
	void Draw_ScoreSheet ( sf::RenderWindow &window ); // 棋譜を表示する
	void Draw_InterfereModePanel ( sf::RenderWindow &window ); // 口出しモードの進行パネルを表示するかもしれない

	void BattleStart();
	void BattleEnd();
	void BattleEnd_Forcebly(); // 強制終了。棋譜は出力しない。つまり試合中にロードする場合の終了処理ということ。

	void AstralProjectionStart ();
	void AstralProjectionEnd ();
	void AstralProjectionEnd_Forcebly (); // 強制終了。つまり検討中にロードする場合の終了処理ということ


	void SetActive ( bool arg ) { _isActive = arg; }
	bool IsActive ()            { return _isActive; }


	void Move ( AI_Modules::Move &move, bool isUseEffect = true  );
	void Move ( NotXML::Tag &moveTag );

	void AddPiece ( string color, string player );
	void AddPiece ( string color, int x, int y );
	void AddCard  ( string cardName, string direction, string player );
	void AddCard  ( string cardName, string direction, int x, int y );

	void RemovePiece ( string player );
	void RemovePiece ( int x, int y );
	void RemoveCard  ( string cardName, string player );
	void RemoveCard  ( int x, int y );

	void RemoveAllPiece_fromBox   ( string player );
	void RemoveAllPiece_fromBoard ();
	void RemoveAllCard_fromBox    ( string player );
	void RemoveAllCard_fromBoard  ();

	void SetFreeOperation ( bool arg );
	bool IsFreeOperation  ();
	
	void RearrangeAllPiecesAndCards ( AI_Modules::InnerExpression *innerExpression );


	///// 棋譜を読み込んでください。
	void ReadScoreSheetFile ( string fileName );

	///// 棋譜の現在位置を指定：検討モード用、ということだと思います。
	void SetScoreIndex ( string index );
					
					
	///// ルールにそった操作を開始。なんかちがう気がするけど、追々修正していきましょう。
	void StartRuleOperation ( int firstTurn );

	///// 最終着手をください。
	AI_Modules::Move GetLastMove();

	///// 何手目かをください。
	int GetScoreRecordLength();

	///// 盤の中心の表示位置を設定。これに合わせて他のパーツが並ぶ。
	void SetBoardCenterPosition ( int x, int y );

	///// 試合のタイトルを設定
	void SetBattleTitle ( string battleTitle );


	///// 操作管理データ関連
	void SetOperation_Lock ( bool arg ) { _operationData.isLock = arg; }
	
	///// 手番の順番
	void SetTurnOrder ( vector<string> turnOrderIdList ) { 

		_operationData.turnOrder = turnOrderIdList; 
			
		// 最初の参加者の手番としておく
		_operationData.SetTurnToFirst();
		/*
		_operationData.currentTurn = "";
		if ( _operationData.turnOrder.size() > 0 ) {
			_operationData.currentTurn = _operationData.turnOrder[0];
		}
		*/
	} 
	/*
	///// 今を、先手の番にする / 後手の番にする
	void SetOperation_CurrentTurn_FIRST  () { _operationData.currentTurn = "FIRST"; }
	void SetOperation_CurrentTurn_SECOND () { _operationData.currentTurn = "SECOND"; }

	///// 先手は P1？ P2？
	void SetOperation_FirstSide_is_P1 () { _operationData.first = "P1"; }
	void SetOperation_FirstSide_is_P2 () { _operationData.first = "P2"; }

	///// 後手は P1？ P2？
	void SetOperation_SecondSide_is_P1 () { _operationData.second = "P1"; }
	void SetOperation_SecondSide_is_P2 () { _operationData.second = "P2"; }

	///// P1 を操作するのは 人間？ AI？
	void SetOperation_P1_is_HUMAN () { _operationData.operatorP1 = "HUMAN"; }
	void SetOperation_P1_is_AI    () { _operationData.operatorP1 = "AI"; }

	///// P2 を操作するのは 人間？ AI？
	void SetOperation_P2_is_HUMAN () { _operationData.operatorP2 = "HUMAN"; }
	void SetOperation_P2_is_AI    () { _operationData.operatorP2 = "AI"; }
	*/

	///// AI設定
	//void SetAIFile                         ( string AI_Name, string fileName );
	//void SetActorScriptFile                ( string AI_Name, string fileName, string playingSide, GUI_SFML::TextMessageWindow *textWindow );
	void SetAIFile                         ( string participantID, string aiFileName );
	void SetActorScriptFile                ( GameManager::Participant participantInfo, string actorFileName, int posX, int posY, int startPosX, int startPosY, GUI_SFML::TextMessageWindow *textWindow );
	
	void SetActorPosition ( int x, int y );
	
	void addParticipant ( Participant participant );
	void clearAllParticipants ();
	
	
	
	
	///// アクターさんにアクセスしたい！
	//Actor* GetActor( string AI_Name );

	///// 現在のターンの参加者を取得
	GameManager::Participant GetCurrentTurnParticipant ();
	
	///// 現在のターンのアクターさんを取得
	Actor* GetCurrentTurnActor ();
	
	///// 全アクターさんの active をセット
	void SetAllActorsActivity ( bool isActive );
	
	///// 全アクターさんを Init
	void InitAllActors ();
	
	///// 全アクターさんを取得
	vector<Actor*> GetAllActors();
	
	///// 現在のターンの参加者の手番フレーム経過数を「開始直後」ではないものにする
	void CurrentTurnParticipant_TurnFrameCountStart();

	///// 現在のターンの参加者の手番フレーム経過数を進める
	void CurrentTurnParticipant_ProgressTurnFrameCount();
	
	///// 現在のターンの参加者の手番フレーム経過数をリセットする
	void CurrentTurnParticipant_ResetTurnFrameCount();
	
	
	/*
	void SetScriptFile_MoveEffectDecision  ( string AI_Name, string fileName );
	void SetScriptFile_EmotionProcess      ( string AI_Name, string fileName );
	void SetScriptFile_Reaction            ( string AI_Name, string fileName );
	string GetScriptFile_MoveEffectDecision  ( string AI_Name );
	string GetScriptFile_EmotionProcess      ( string AI_Name );
	string GetScriptFile_Reaction            ( string AI_Name );
	*/

	///// 試合関係
	bool IsBattlingNow() { return _battleInfo.isBattlingNow; }
	string GetBattleResult () { return _battleInfo.result; }
	string GetBattleResultDescription () { return _battleInfo.resultDescription; }


	///// AIは思考を完了している？
	bool IsAIThinkFinished();

	///// AIさんの思考結果をください。結果の方。
	AI_Modules::ProgressData GetProgressData_AfterThinking() { return _progressData_AfterThinking; }

	///// AIさん、着手を実行してね：エフェクトレベルを指定（0〜100値。内部で、用意しているエフェクト数に換算）
	void ExecuteAIMove ( int effectLevel );
	void ExecuteAIMoveWithCutIn ( IEffect *cutInEffect );
	DelegateBase* GetEndMoveFromAICutIn () { return _endMoveFromAICutIn; }

	///// AIさん着手実行中？
	bool IsMoveEffectGoing() { return _is_MoveEffectGoing; }

	///// 棋譜にアクセスさせてー：Let me get an access to the game record.
	GameRecord* GetGameRecord () { return _score; }

	///// 今の操作データをください。
	OperationData GetOperationData () { return _operationData; }
	
	
	///// ズバリ、カードをつかんでいる？：つまり、右クリックがメニュー表示とカード回転でかぶるので、一時的にメニュー表示を抑制する意図
	bool IsHoldingCard();
	
	///// 棋譜を表示する？ しない？
	void ShowScoreSheetViewer();
	void HideScoreSheetViewer();
	
	bool IsItOkayToBeRendered() { return _isRendered; }
	
	sf::Text _battleTitleForKifuViewing; // 棋譜再生時用の試合タイトルラベル
	void SetupBattleTitleForKifuViewing ( string battleTitle );
	
	
	//=========================================================
	// セーブ/ロード用のインターフェイス
	//=========================================================
	struct SaveData {
		
		string name;
		string gameBoardSerialized;
		string gameRecordSerialized; 
		bool isActive;
		bool isFreeOperation;
		int boardCenterX;
		int boardCenterY;
		
		int actorPosisionX;
		int actorPosisionY;

		bool  isRendered;
		int   operationState;
		bool  IsSelectableMoveListNew;
		bool  isAstralProjection;
		bool  is_MoveEffectGoing;

		string evaluationSumUp_FromAI;

		OperationData operationData;
		BattleInfo    battleInfo;

		sf::Vector2f kifuViewerPosition;
		
		vector<AI_Modules::Move> selectableMoveList;
		
		AI_Modules::ProgressData progressData_AfterThinking;
		
		string astralProjection_physicalState; // あくまでも検討モード用。全データが必要なわけではない。そこのところは GameManager側の都合なので、文字列との変換処理は GameManager側で行うこととする　AI_Modules::InnerExpression *_astralProjection_physicalState
		
		// 負の数なら選択中のグリッドはないものとする。
		int selectingGridX;
		int selectingGridY;
		
		
		string operationMemo_belongsToWhichPieceBox; // マウスで持ち上げた駒が、どこの駒台のものであったのか？ P1 or P2
		int    operationMemo_whatColorThePieceWas;   // マウスで持ち上げた駒が、何色であったのか？ Piece::Color::BLACK or WHITE
		string operationMemo_belongsToWhichCardBox;
		string operationMemo_whichCardTheCardWas;
		
		
		//string AI_fileName;
		//string actorSerialized;
		
		string participantsString;
		
		string battleTitle;
		
	};
	
	SaveData GetSaveData();
	void Load ( GameManager::SaveData &saveData );
	
	
	// 盤上に仮表示する指し手。
	AI_Modules::Move _moveForDemonstration;
	
	void SetMoveDemonstration   ( AI_Modules::Move moveForDemonstration ) { _moveForDemonstration = moveForDemonstration; }
	void ClearMoveDemonstration () { _moveForDemonstration.action = 0; }
	
	
	///// 口出しモード関連
	void SetInterferenceModeInfo ( InterferenceModeInfo arg ) { _interferenceModeInfo = arg; }
	InterferenceModeInfo GetInterferenceModeInfo () { return _interferenceModeInfo; }
	
	//// 口出しモード進行
	void Update_InterferenceMode ( AI_Modules::ProgressData AI_thinkResult );
	
	// 口出しモード：AIの指し手候補待ち受け開始
	void InterferenceMode_StartWaitingForAIMove ();
	
	// 口出しモード：プレイヤーが口出し案を決定した
	void InterferenceMode_PlayerMoveDecided ( AI_Modules::Move moveByPlayer );
	
	// 口出しモード管理者さんへの着手完了通知：手番に関係なく（手番交代前に）
	void InterferenceMode_ModeDone ();

	// 口出しモード管理者さんへの手番開始通知：手番に関係なく（手番交代前に）
	void InterferenceMode_NewTurnStarts();
	
	// 口出しモードでのプレイヤー操作中であるかどうか
	bool InterferenceMode_IsPlayersDecision ();
	
	// 口出しモードとして処理しますか？（ScriptEngine側での進行処理の問題）
	bool IsInterferenceMode ();

	// 口出しモードの試合ですか？（試合自体が口出しモードであるか否か）
	bool IsInterferenceBattle ();

	// 現在の手番の参加者は口出され対象ですか？
	bool InterferenceMode_IsCurrentParticipantIsSubjectOfInterferenceMode ();

	///// 指定の参加者の口出しモードを設定する
	void SetInterferenceModeToParticipant ( string participantId, int maxTimes, string selectOptionsPos );
	
	///// 参加者の口出しモード情報をGamaManager側に展開する
	void DeployInterferenceModeFromParticipant ( string participantId );

	///// 参加者の口出しモード情報に、GamaManager側から保存する
	void WriteInterferenceModeToParticipant ( string participantId );
	
	
	
	
	
	
	
	///// 指定の参加者の候補手からの選択モードの設定をする
	void SetChoiceModeToParticipant ( string participantId, int numberOfChoices, string selectOptionsPos, vector< pair<int, int>> extraChoicesRange );

	///// 参加者の候補手からの選択モード情報をGamaManager側に展開する
	void DeployChoiceModeFromParticipant ( string participantId );

	///// 参加者の口候補手からの選択モード情報に、GamaManager側から保存する
	void WriteChoiceModeToParticipant ( string participantId );
	
	
	// 候補手からの選択モードとして処理しますか？（ScriptEngine側での進行処理の問題）
	bool IsChoiceMode ();
	
	//// 候補手からの選択モード進行
	void Update_ChoiceMode ( AI_Modules::ProgressData AI_thinkResult );

	// 手番交代後の、候補手からの選択モードの進行処理（新手番の人が対象者ならば云々……）
	void ChoiceMode_NewTurnStarts();
	
	///// 候補手からの選択モードであれば、プレイヤーが選んだものに置き換えたものを返す
	AI_Modules::ProgressData ReplaceToPlayersChoiceIfNeeded ( AI_Modules::ProgressData AI_think_result );
	
	//=========================================================
	// 現在の手番表示パネルに現在の手番の名前を設定
	//=========================================================
	void RenewCurentTurnNamePanel ();
	
	//=========================================================
	// 掛け合い受付
	//=========================================================
	void AcceptKakeai ( Actor::KakeaiInfo kakeaiInfo );
	//void AcceptKakeai ( string participantId, string remark, string faceName );
	
	//=========================================================
	// ペアバトルのチームを認識。Participantsタグの処理の終わりに。
	//=========================================================
	void TeamForming();
	
	//=========================================================
	// ペアバトルのチームに対して、ターンの開始を告げる。表示位置などをよしなに
	//=========================================================
	void TeamNewTurnStart();
        
    
    vector<Piece*> GetAllPieces_forManagement();
    vector<Card*>  GetAllCards_forManagement();
    
    bool _isScriptingMoving;
    void StartScriptingMoving();
    bool IsScriptingMovingFinished();
    
    
    void StartTutorialEvent() { _isTutorial = true; }
    
    
    GameRecord* GetScoreRecorder() { return _score;} // ScriptEngine側で「終局」を記録したいということ
    
private:
	
	///// 試合中じゃないけど、着手イベントの待ち受けをしたい場合のフラグ。つまりチュートリアルか。
	bool _isTutorial;
	
	ScriptEngine *_scriptEngine;
	
	
	///// 盤上に指し手を仮表示する関連
	void DrawMoveDemonstration ( sf::RenderWindow &window );
	Arrow *_arrowForMoveDemonstration;
	Image *_crashForMoveDemonstration;
	
	///// 口出しモード関連
	InterferenceModeInfo _interferenceModeInfo;
	Select               *_select;
	
	///// 口出しモードでプレイヤーが着手を決定した
	//void PlayerMoveDone_InterferenceMode();

	///// 候補手からの選択モード関連
	ChoiceModeInfo _choiceModeInfo;
	
	///// 千日手関連
	//
	// プレイヤーが着手しようとするタイミングでの（←重要）千日手判定
	int CountRepetition ( AI_Modules::Move attemptingMove );
	
	///// ペアバトル関連
	// 各サイドのチームのアクター
	vector<Actor*> _teamActorP1;
	vector<Actor*> _teamActorP2;
	
	
	///// セーブデータ作成/復元
	string SerializeGameBoard();
	void   DeSerializeGameBoard( string data );
	
	///// Participantのシリアライズ/デシリアライズ
	string      SerializeParticipant ( Participant participant );
	Participant DeSerializeParticipant( string data );
	
	
	
	
	//=========================================================
	// 内部定数：ルール上の操作状態
	//=========================================================
	struct OperationState {
		static const int TO_DETERMIN_WHAT_TO_DO = 0; // まだ何をするか決めていない
		static const int TO_PUT_PIECE           = 1; // 駒台上の駒を、盤面に、設置しようとしている
		static const int TO_MOVE_PIECE          = 2; // 盤面上の駒を、移動させようとしている
		static const int TO_PUT_CARD            = 3; // カード置き場のカードを、盤面に、設置しようとしている
	};
	
	string _name;
	
	///// 描画してもいい？
	bool _isRendered;

	///// ルール上の、操作状態
	int _operationState;

	///// 手番
	OperationData _operationData;
	//int _currentTurn;

	///// 現在、可能な指し手、のリスト
	vector<AI_Modules::Move> _selectableMoveList;
	bool _IsSelectableMoveListNew; // が、最新かどうか。

	///// 駒移動時、選択中のグリッド
	Grid *_selectingGrid;
		// セーブ/ロード用に盤上の位置情報を覚えておく
		int _selectingGridX;
		int _selectingGridY;

	bool _isFreeOperation;
	
	bool _isAstralProjection;
	AI_Modules::InnerExpression *_astralProjection_physicalState;

	bool _isActive;

	Board    *_board;
	CardBox  *_cardBox_P1;
	CardBox  *_cardBox_P2;
	PieceBox *_pieceBox_P1;
	PieceBox *_pieceBox_P2;

	vector<Piece*> _pieceList_forManagement;
	vector<Card*>  _cardList_forManagement;  
		// つまり、検討モード（フリー操作中）でマウスを持ったまま終了したときのためのメモ
		/*
			マウスにくっついているカードを外すにしても、
			「このGameManagerに由来するものである」という保証が欲しい。
			そのためだけのメモ。
		*/

	// AI思考ゲージ
	GUI_SFML::ProgressBar *_AI_ProgressBar;

	// 気力ゲージ：AIにはAI思考ゲージ、プレイヤーには気力ゲージ、ということでどうだ！（つまり P1/P2両方の分は持たない）
	PsychoGage *_psychoGage;
	
	// 口出しモードの状況パネル
	InterfereModePanel * _interfereModePanel;
	
	// 現在の手番表示パネル
	CurrentTurnNamePanel *_currentTurnNamePanel;

	//=========================================================
	// 各パーツの表示位置
	//=========================================================
	float _boardX;
	float _boardY;
	float _carBoxP1X;
	float _carBoxP2X;
	float _cardBoxP1Y;
	float _cardBoxP2Y;
	float _pieceBoxP1X;
	float _pieceBoxP2X;
	float _pieceBoxP1Y;
	float _pieceBoxP2Y;
	float _AI_ProgresBarX_P1;
	float _AI_ProgresBarY_P1;
	float _AI_ProgresBarX_P2;
	float _AI_ProgresBarY_P2;
	float _interfereModePanelX;
	float _interfereModePanelY;

	float _boardCenterX;
	float _boardCenterY;
	
	// アクターの位置（反転を反映させる前の基準位置。反転させない場合はこのまま。反転させる場合はXを反転）
	/*
		むしろP1の場合に、位置が反転することとなる。
	*/
	//float _actorPosisionX;
	//float _actorPosisionY;

	//=========================================================
	// 各パーツの表示位置を定める
	//=========================================================
	void SetUpPartsDrawPositions();

	//=========================================================
	// //AIをセットアップ
	//=========================================================
	//void SetUpAI();

	//=========================================================
	// ルールをセットアップ
	//=========================================================
	void SetUpRule();

	//==========================================================
	// 駒オブジェクトを解放
	//==========================================================
	void ReleaseAllPiece();
	void ReleasePiece ( Piece *piece );

	//=========================================================
	// 駒台に駒を乗せる
	//=========================================================
	void AddPieceToPieceBox ( PieceBox *pieceBox, int pieceNum, int color );



	//=========================================================
	// 指し手をGUI上の盤面で実行する
	//=========================================================
	void ExecuteMove( const AI_Modules::Move &move, bool isUseEffect = true );

	//=========================================================
	// 現在の局面を内部表現化する
	//【引数】
	// 今、この手番だとして。AI_Module の TURN::BLACK / WHITE
	//=========================================================
	AI_Modules::InnerExpression* GenerateInnerExpressionFromCurrentState();
	AI_Modules::InnerExpression* GenerateInnerExpressionFromCurrentState( int turn );


	//=========================================================
	// 全カードを取得：盤面＆両カード置き場
	//=========================================================
	vector<Card*> GetAllCards();


	//=========================================================
	// ルールにそった操作を受付
	//=========================================================
	void Operation_With_Rule ();
	void Operation_With_Rule_TO_DETERMIN_WHAT_TO_DO ();
	bool Operation_With_Rule_TO_DETERMIN_WHAT_TO_DO__MAYBE_PUT_PIECE  ();
	bool Operation_With_Rule_TO_DETERMIN_WHAT_TO_DO__MAYBE_MOVE_PIECE ();
	bool Operation_With_Rule_TO_DETERMIN_WHAT_TO_DO__MAYBE_PUT_CARD   ();

	void Operation_With_Rule_TO_PUT_PIECE           ();
	void Operation_With_Rule_TO_MOVE_PIECE          ();
	void Operation_With_Rule_TO_PUT_CARD            ();
	
	// 人間の指し手が決定したタイミングで指し手情報を受け取って何かするかもしれない
	void humanMoveDecided ( AI_Modules::Move decidedMove );

	//=========================================================
	// ルールにそって、指し手を実行した後の事後処理
	//=========================================================
	void Operation_AfterExecute_Process ( AI_Modules::Move move );

	//=========================================================
	// グリッドのフラッシュを消す
	//=========================================================
	void GridsFlashCancel();

	//=========================================================
	// 選択可能なグリッドをフラッシュさせる
	//=========================================================
	void FlashSelectableGrids_ToPutPiece  ();
	void FlashSelectableGrids_ToMovePiece ( int selectedX, int selectedY );
	void FlashSelectableGrids_ToPutCard   ( int cardId, int direction );
	
	///// フラッシュを解除するときには、指し手の指定は不要のはず。
	void FlashEndAllGrids ();

	//=========================================================
	// 現在の局面で可能な指し手が最新じゃないなら作成する
	//=========================================================
	void MakeSelectableList_IfNeccesary ();


	///// AI思考関連
	//AI_Modules::AI *_AI;
	AI_Manager     *_AI_Manager;
	//Actor          *_actor;
	
	vector<Participant> _participantsList;


	
	// セーブ/ロード用のメモ
	string _AI_fileName;

	///// ファイル名ではなく、読み込み後の Luaオブジェクトを持つことにすれば、内部の状態を保っておける？
	//string _scriptFile_MoveEffectDecision; // 外部スクリプトファイル：着手時のエフェクト決定処理
	//string _scriptFile_EmotionProcess;     // 外部スクリプトファイル：感情値変動処理
	//string _scriptFile_Reaction;           // 外部スクリプトファイル：リアクション処理



	//=========================================================
	// 試合関連
	//=========================================================
	///// 棋譜？
	//vector < AI_Modules::Move > _score;
	GameRecord                  *_score;
	sf::Vector2f                 _kifuViewerPosition;

	///// 試合情報
	BattleInfo _battleInfo;

	///// 勝敗判定：着手のたびに呼び出す。
	void Judgement();

	///// この色、勝ってる？
	bool IsWin ( AI_Modules::InnerExpression *ie, int color );

	///// 今、AIの番？
	bool Is_AI_Turn_Now();

	///// 試合中で、どっちかがAI？
	bool Is_AI_Participating_Now();

	///// 着手エフェクトオブジェクトに渡す終了通知受け取りメソッド
	void EndMoveFromEffect ( AI_Modules::Move move );
	DelegateBase *_endMoveFromEffect;
	DelegateBase *_endMoveFromAICutIn; // カットイン終了時に呼んで欲しいもの。ExecuteMove(int effectLebel)

	///// 着手している最中。エフェクト終了待ち。
	bool _is_MoveEffectGoing;

	///// 思考完了後の成果物：スクリプトエンジン経由で、実行されることを待っている指し手
	AI_Modules::ProgressData _progressData_AfterThinking;

	///// 思考完了後の成果物：指し手の総括評価。棋譜記録用に、指し手エフェクトの前に退避
	string _evaluationSumUp_FromAI;

	//=========================================================
	// 現在の手番は、AI の TURN としてはどっちか？
	// AI::TURN の 黒・白に、GameManager のFirst・Second を変換
	//=========================================================
	int GetTurn_in_tarms_of_TURN_FOR_AI ();

	//=========================================================
	// 現在の手番の駒台をください
	//=========================================================
	PieceBox* GetPieceBox_CurrentTurn();

	//=========================================================
	// 現在の手番のカード置き場をください
	//=========================================================
	CardBox* GetCardBox_CurrentTurn();

	//=========================================================
	// 現在の手番じゃない方の駒台をください
	//=========================================================
	PieceBox* GetPieceBox_NotCurrentTurn();

	//=========================================================
	// この駒が属している方の駒台をください
	//=========================================================
	PieceBox* GetPieceBox_BelongToPiece ( Piece *piece  );

	//=========================================================
	// ぶつける先のグリッドが欲しい。
	//【引数】from to、では、その先のグリッドは？
	//=========================================================
	Grid* GetCollidedGrid ( int fromX, int fromY, int toX, int toY );
	
	
	//=========================================================
	// 検討モードでの再生用。盤面が一致しているかどうか。
	// つまり、「次へ」ボタンで駒などを動かすとき、フリー操作で盤面がいじられた状態かどうかを判定したい。
	// InnerExpression の IsEqual() ほど厳密に同一性を調査する必要はない 
	//=========================================================
	bool IsEqual_EasyVersion ( AI_Modules::InnerExpression *ie1, AI_Modules::InnerExpression *ie2 );
	
	
	//=========================================================
	// 検討モードでの盤面いじり判定用のデータ。セーブロード対応のための処理
	//=========================================================
	string ToString_astralProjection_physicalState ();
	void FromString_astralProjection_physicalState ( string fromSaveData );
	
	
	//=========================================================
	// つまり、ロード時に、操作用のエフェクトを削除する目的
	//=========================================================
	vector<IEffect*> _takeCareOfTheseEffects;
	void LetsRemenberThatIMadeThisEffect ( IEffect *effect );
	void TerminateAllOurBeLovedEffects();
	
	//=========================================================
	// 盤面を操作中にセーブした、ですって？ 面倒なことしないでくださいよね、もうー。
	//=========================================================
	string _operationMemo_belongsToWhichPieceBox; // マウスで持ち上げた駒が、どこの駒台のものであったのか？ P1 or P2
	int    _operationMemo_whatColorThePieceWas;   // マウスで持ち上げた駒が、何色であったのか？ Piece::Color::BLACK or WHITE
	string _operationMemo_belongsToWhichCardBox;
	string _operationMemo_whichCardTheCardWas;
	
	//=========================================================
	// うちの子をマウスでつかんでるなら、マウスから外して解放する
	//=========================================================
	void ReleaseCardFromMouseIfBelongToHere();
	void ReleasePieceFromMouseIfBelongToHere();
	

	//=========================================================
	// 指し手が同じかどうか（千日手判定の目的で）
	//=========================================================
	bool IsEqualMove ( AI_Modules::Move move1, AI_Modules::Move move2 );
	
	//=========================================================
	// 指し手の簡易文字列表現
	//=========================================================
	string MoveToString ( AI_Modules::Move &mv );
	
	

	//=========================================================
	// 現局面の簡易スクリプト化
	//=========================================================
	string BoardToString();
	
	
	
	
#if defined(_WIN32)
	
	//=========================================================
	// 棋譜ファイルが増えた場合にフォルダにまとめる処理
	//=========================================================
	void ArchiveRecordFiles ();

#elif defined(__linux__)


#endif
};





#endif

